package be.iminds.ilabt.jfed.espec.filefetcher;

import be.iminds.ilabt.jfed.espec.bundle.Bundle;
import be.iminds.ilabt.jfed.espec.model.ExecuteSpec;
import be.iminds.ilabt.jfed.espec.model.ExperimentSpecification;
import be.iminds.ilabt.jfed.espec.model.FileSource;
import be.iminds.ilabt.jfed.espec.model.UploadSpec;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * ExperimentSpecificationFileManager
 *   Fetches all files in an ExperimentSpecification, putting them in memory.
 *   It downloads any files needing download in parallell,
 *   and has a method to wait for all files to be ready.
 */
public class ExperimentSpecificationFileManager {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentSpecificationFileManager.class);

    @Nonnull private final ExperimentSpecification eSpec;
    @Nonnull private final Bundle bundle;

    private final Map<FileSource, byte[]> bytes = new ConcurrentHashMap<>();
    private final ExecutorService executorService;

    private boolean started;
    private boolean errorOccured;

    public ExperimentSpecificationFileManager(@Nonnull ExperimentSpecification eSpec, @Nonnull Bundle bundle) {
        this.eSpec = eSpec;
        this.bundle = bundle;
        this.executorService = Executors.newFixedThreadPool(3);
        this.errorOccured = false;
    }

    public void fetchAll() {
        started = true;

        Map<Boolean, List<FileFetcher>> fileSourceListBySpeed =
                Stream.concat(
                        eSpec.getUploads().stream().map(UploadSpec::getSource),
                        eSpec.getExecutes().stream().map(ExecuteSpec::getSource)
                ).map(fileSource -> new FileFetcher(fileSource, bundle))
                        .collect(Collectors.partitioningBy(FileFetcher::isFast));

        List<FileFetcher> fastFetchers = fileSourceListBySpeed.get(true);
        List<FileFetcher> slowFetchers = fileSourceListBySpeed.get(false);

        for (FileFetcher fff : fastFetchers) {
            try {
                byte[] b = fff.fetchBytes();
                assert b != null;
                LOG.debug("ExperimentSpecificationFileManager fetched "+b.length+" bytes for "+fff.getFileSource().getBasename());
                bytes.put(fff.getFileSource(), b);
            } catch (IOException e) {
                LOG.error("Failed to (fast) fetch file with source type "+fff.getFileSource().getType()+"", e);
                errorOccured = true;
                executorService.shutdown();
                return;
            }
        }

        LOG.debug("ExperimentSpecificationFileManager fetched "+fastFetchers.size()+" fast files.");

        for (FileFetcher sff : slowFetchers) {
            queueFileFetch(sff);
        }

        LOG.debug("ExperimentSpecificationFileManager queued "+slowFetchers.size()+" slow files for fetching.");

        executorService.shutdown();
    }

    private void queueFileFetch(@Nonnull FileFetcher fileFetcher) {
        Runnable backGroundFetch = new Runnable() {
            @Override
            public void run() {
                try {
                    bytes.put(fileFetcher.getFileSource(), fileFetcher.fetchBytes());
                } catch (IOException e) {
                    LOG.error("Failed to (slow) fetch file with source type "+
                            fileFetcher.getFileSource().getType()+"", e);
                    errorOccured = true;
                    executorService.shutdownNow();
                }
            }
        };
        executorService.execute(backGroundFetch);
    }

    public boolean isReady() {
        if (!started)
            throw new IllegalStateException("Dangerous to call isReady before starting!");
        return executorService.isTerminated();
    }

    public boolean getErrorOccured() {
        return errorOccured;
    }

    /**
     * Wait for up to 30 minutes for all tasks to become ready
     *
     * @throws InterruptedException
     */
    public void waitUntilReady() throws InterruptedException {
        if (!started)
            throw new IllegalStateException("Dangerous to call waitUntilReady before starting!");

        if (isReady()) {
            LOG.debug("ExperimentSpecificationFileManager did not have to wait for all files to be ready");
            return;
        }

        LOG.debug("ExperimentSpecificationFileManager waits for all files to be ready");

        boolean terminated = this.executorService.awaitTermination(30, TimeUnit.MINUTES);
        if (!terminated)
            throw new IllegalStateException("Failed to wait until all files are ready");

        LOG.debug("ExperimentSpecificationFileManager stopped waiting for files to be ready ready");
    }

    @Nonnull
    public byte[] getBytes(@Nonnull FileSource fileSource) {
        if (!isReady())
            throw new IllegalStateException("ExperimentSpecificationFileManager requires you to wait until all files are ready, before using them.");

        return bytes.get(fileSource);
    }




    @Nonnull
    public ExperimentSpecification getExperimentSpecification() {
        return eSpec;
    }

    @Nonnull
    public Bundle getBundle() {
        return bundle;
    }
}
