package be.iminds.ilabt.jfed.highlevel.jobs.states;

import be.iminds.ilabt.jfed.espec.bundle.Bundle;
import be.iminds.ilabt.jfed.espec.model.ExecuteSpec;
import be.iminds.ilabt.jfed.espec.model.ExperimentSpecification;
import be.iminds.ilabt.jfed.espec.model.UploadLikeSpec;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.jobs.Job;
import be.iminds.ilabt.jfed.highlevel.jobs.SlicedState;
import be.iminds.ilabt.jfed.highlevel.jobs.State;
import be.iminds.ilabt.jfed.highlevel.jobs.StateSlice;
import be.iminds.ilabt.jfed.highlevel.util.ByteArraySshjFile;
import be.iminds.ilabt.jfed.highlevel.util.ExternalFileUtil;
import be.iminds.ilabt.jfed.highlevel.util.ProxySocketFactoryProvider;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.rspec.model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.DistributeSshKeypair;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.util.BestNodeLoginFinder;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.transport.TransportException;
import net.schmizz.sshj.transport.verification.PromiscuousVerifier;
import net.schmizz.sshj.userauth.UserAuthException;
import net.schmizz.sshj.userauth.keyprovider.KeyPairWrapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.net.SocketFactory;
import java.io.IOException;
import java.security.KeyPair;
import java.util.*;

/**
 * ExperimentSpecificationServicesDirsAndUploadsState does 3 things:
 *   - Setup SSH connections to nodes, for the dirs, upload, and execute steps
 *   - Do the "dir" step of the ExperimentSpecification
 *   - Do the "upload" step of the ExperimentSpecification
 *
 * Note that these 2 steps can be combined here easily, as there are no inter node dependencies.
 * For the execute step, things are different. See ExperimentSpecificationServicesSingleExecuteState
 *
 * Additionally, this class can be called to close all SSH connections. (typically after all executes)
 *   See TODO
 */
public class ExperimentSpecificationServicesDirsAndUploadsState extends SlicedState<ExperimentSpecificationServicesDirsAndUploadsState.ExperimentSpecificationServiceDirsAndUploadsStateSlice> {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentSpecificationServicesDirsAndUploadsState.class);
    public static final int MAX_EXECUTE_FINISHED_FILE_DETECTION_TRIES = 60;
    public static final int TIME_BETWEEN_FINISHED_FILE_DETECTION_TRIES_MS = 10_000;

    @Nonnull private final Job<?> job;
    @Nonnull private final Experiment experiment;
    @Nonnull private final GeniUserProvider geniUserProvider;
    @Nonnull private final ProxySocketFactoryProvider proxySocketFactoryProvider;

    @Nonnull private final ExperimentSpecification experimentSpecification;
    @Nonnull private final Bundle experimentSpecificationBundle;

    @Nonnull
    private final KeyPair loginKeyPair;


    @Nonnull private final List<ExperimentSpecificationServiceDirsAndUploadsStateSlice> slices;

    @Nullable private final BestNodeLoginFinder bestNodeLoginFinder;

    @Nonnull private final Map<String, SSHClient> sshClientByClientId = new HashMap<>();
    @Nonnull private final Map<UploadLikeSpec, String> remotePathByUploadLikeSpec = new HashMap<>();

    protected ExperimentSpecificationServicesDirsAndUploadsState(@Nonnull Job<?> job,
                                                                 @Nonnull GeniUserProvider geniUserProvider,
                                                                 @Nonnull ProxySocketFactoryProvider proxySocketFactoryProvider) {
        super("Execute ExperimentSpecification Dir And Upload Steps");
        this.job = job;
        this.experiment = job.getExperiment();
        this.geniUserProvider = geniUserProvider;
        this.proxySocketFactoryProvider = proxySocketFactoryProvider;

        if (experiment.getExperimentSpecification() == null)
            throw new RuntimeException("Internal error, ExperimentSpecificationServicesDirsAndUploadsState created for experiment without ExperimentSpecification");
        if (experiment.getExperimentSpecificationBundle() == null)
            throw new RuntimeException("Internal error, ExperimentSpecificationServicesDirsAndUploadsState created for experiment without ExperimentSpecification (bundle)");

        this.experimentSpecification = experiment.getExperimentSpecification();
        this.experimentSpecificationBundle = experiment.getExperimentSpecificationBundle();

        assert experiment.getSlice() != null;

        if (experiment.getSlice() == null)
            throw new IllegalStateException("Experiment should have a slice");

        assert experiment.getKeypairs() != null;

        String distributedSshKeysUsername = null;
        KeyPair loginKeyPair = null;

        for (Map.Entry<DistributeSshKeypair, KeyPair> e : experiment.getKeypairs().entrySet()) {
            if(e.getValue() != null){
                //distributedSshKeysUsername may be null, but that's ok, that just means the logged in users account needs to be used.
                distributedSshKeysUsername = e.getKey().getUser() == null ? null : e.getKey().getUser().getResourceName();

                loginKeyPair = e.getValue();
                break;
            }
        }

        if(loginKeyPair == null)
            loginKeyPair = new KeyPair(geniUserProvider.getLoggedInGeniUser().getPublicKey(),
                    geniUserProvider.getLoggedInGeniUser().getPrivateKey());

        this.loginKeyPair = loginKeyPair;


        this.slices = new ArrayList<>();

        RequestRspecSource requestRspecSource = experiment.getSlice().getRequestRspec();
        FXModelRspec model = requestRspecSource == null ?
                null :
                requestRspecSource.getModelRspec(ModelRspecType.FX);
        assert model != null;

        if (model == null) {
            LOG.warn("ExperimentSpecificationServicesDirsAndUploadsState will do nothing because there is no usable request RSpec");
            bestNodeLoginFinder = null;
            return;
        }

        for (FXRspecNode node : model.getNodes()) {
            String nodeId = node.getClientId();

            if (nodeId != null) {
                //Note: Since this needs to setup SSH connections for execute as well, this will also run on nodes with only an execute step.

                boolean neededOnNode = experimentSpecification.getDirs().stream().anyMatch(s -> s.getNodes() == null || s.getNodes().contains(nodeId));
                neededOnNode |= experimentSpecification.getExecutes().stream().anyMatch(s -> s.getNodes() == null || s.getNodes().contains(nodeId));
                neededOnNode |= experimentSpecification.getUploads().stream().anyMatch(s -> s.getNodes() == null || s.getNodes().contains(nodeId));

                if (neededOnNode) {
                    slices.add(new ExperimentSpecificationServiceDirsAndUploadsStateSlice(job, node));
                }

                //TODO: we need to make sure we close all ssh connections on any error
                //TODO:   we proably just need a state that closes them all for this.
            }
        }

        ManifestRspecSource manifestRspecSource = experiment.getSlice().getManifestRspec();
        bestNodeLoginFinder = manifestRspecSource == null ||
                manifestRspecSource.getStringRspec() == null ? null :
                new BestNodeLoginFinder(manifestRspecSource.getStringRspec(),
                        distributedSshKeysUsername, this.geniUserProvider.getLoggedInGeniUser(),
                        new BestNodeLoginFinder.Feedback() {
                            @Override
                            public void info(String s) {
                                LOG.info("ExperimentSpecificationServicesDirsAndUploadsState: " + s);
                            }

                            @Override
                            public void error(String s) {
                                LOG.error("ExperimentSpecificationServicesDirsAndUploadsState: " + s);
                            }

                            @Override
                            public void debug(String s) {
                                LOG.debug("ExperimentSpecificationServicesDirsAndUploadsState: " + s);
                            }
                        });
    }

    private void closeAllSshConnections() {
        for (SSHClient ssh : sshClientByClientId.values()) {
            try {
                if (ssh != null)
                    ssh.close();
            } catch (IOException ignored) {
                //probably, connection is already closed. We don't really need to worry about this.
                LOG.debug("Will ignore error while closing SSH connection: "+ignored.getMessage());
            }
        }
        //make sure it isn't reused
        sshClientByClientId.clear();
    }

    @Override
    public Collection<ExperimentSpecificationServiceDirsAndUploadsStateSlice> getSlices() {
        return slices;
    }

    @Nullable
    public SSHClient getSshClient(@Nonnull String clientId) {
        return sshClientByClientId.get(clientId);
    }

    @Nullable
    public String getRemotePath(@Nonnull UploadLikeSpec uploadLikeSpec) {
        return remotePathByUploadLikeSpec.get(uploadLikeSpec);
    }


    public class AssureAllSshClosedStateSlice extends State {
        protected AssureAllSshClosedStateSlice() {
            super("Close All SSH Connections");
        }

        @Nonnull
        @Override
        protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
            closeAllSshConnections();
            return ExperimentTaskStatus.SUCCESS;
        }
    }

    public AssureAllSshClosedStateSlice getAssureAllSshClosedStateSlice() {
        return new AssureAllSshClosedStateSlice();
    }

    public class ExperimentSpecificationServiceDirsAndUploadsStateSlice extends StateSlice {
        @Nonnull protected final FXRspecNode node;
        @Nonnull protected final String clientId;
        @Nullable  private SSHClient ssh;

        @Nullable private ExternalFileUtil externalFileUtil;

        public ExperimentSpecificationServiceDirsAndUploadsStateSlice(@Nonnull Job<?> job,
                                                                      @Nonnull FXRspecNode node) {
            super(job);
            this.node = node;
            if (node.getClientId() == null)
                throw new IllegalStateException("Only nodes with a client_id can be used in an ExperimentSpecification");
            this.clientId = node.getClientId();
        }

        @Override
        public String getName() {
            return "ExperimentSpecification steps for node "+node.getClientId();
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            final SetupSshConnectionState setupSshConnectionState = new SetupSshConnectionState();
            setAndRunState(setupSshConnectionState);

            if (setupSshConnectionState.getStatus() != ExperimentTaskStatus.SUCCESS)
                return ExperimentTaskStatus.FAILED;

            //TODO: implement -> For now skip it? not supported in first version!
//            final SetupDirsState setupDirState = new SetupDirsState();
//            setAndRunState(setupDirState);

            //TODO: implement upload based on uploadInventory
            final UploadState uploadState = new UploadState();
            setAndRunState(uploadState);

//            //TODO: implement ExecuteState based on execute of ansible script
//            final ExecuteState executeState = new ExecuteState();
//            setAndRunState(executeState);
//
//            //TODO: later, we might need something like this for ADVANCED scripts
////            //check if there are finished-files to wait for
////            if (node.getExecuteServices().stream().anyMatch(es -> es.getFinishedFile() != null)) {
////                setAndRunState(new WaitForExecuteServicesState());
////            } else {
////                LOG.info("Did not find any execute-services with finished-file attribute. We assume that all files that are needed are present!");
////            }
            return null;
        }

        public class SetupSshConnectionState extends State {

            protected SetupSshConnectionState() {
                super("Setting up SSH Connection to '" + node.getClientId() + "'");
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                //get login-service associated to the user we are distributing the key to
                BasicStringRspec.LoginService loginService =
                        bestNodeLoginFinder == null ?
                                null :
                                bestNodeLoginFinder.findBestLogin(node.getUniqueId());

                if(loginService == null){
                    LOG.warn("Cannot setup connection to {} as no login-service is available", node.getClientId());
                    return ExperimentTaskStatus.FAILED;
                }

                ssh = new SSHClient();
                ssh.addHostKeyVerifier(new PromiscuousVerifier());

                sshClientByClientId.put(clientId, ssh);


                try {
                    SocketFactory socketFactory = proxySocketFactoryProvider.createProxySocketFactory(loginService.getHostname(), loginService.getPort());
                    if (socketFactory != null) {
                        LOG.debug("Using proxy for ExperimentSpecification");
                        ssh.setSocketFactory(socketFactory);
                    } else {
                        LOG.debug("No proxy needed for ExperimentSpecification");
                    }
                } catch (IOException ex) {
                    LOG.error("Error while creating ProxySocketFactory. Trying without now.", ex);
                }

                try {
                    ssh.connect(loginService.getHostname(), loginService.getPort());
                } catch (IOException e) {
                    updateMessage("Error while connecting to " + loginService.getHostname() + ": " + e.getMessage());
                    LOG.error("Error while connecting to {}: {}", loginService.getHostname(), e.getMessage(), e);
                    return ExperimentTaskStatus.FAILED;
                }
                try {
                    //note: we must use the distributed ssh keys to login: in speaksfor case, there is no other way to login
                    ssh.authPublickey(loginService.getUsername(), new KeyPairWrapper(loginKeyPair));

                    externalFileUtil = new ExternalFileUtil(ssh);
                } catch (UserAuthException e) {
                    updateMessage("Could not authenticate: " + e.getMessage());
                    LOG.error("Authentication failed on {}: {}", node.getClientId(), e.getMessage(), e);
                    return ExperimentTaskStatus.FAILED;
                } catch (TransportException e) {
                    updateMessage("Error while authenticating: " + e.getMessage());
                    LOG.error("Error while authenticating on {}: {}", node.getClientId(), e.getMessage(), e);
                    return ExperimentTaskStatus.FAILED;
                }

                return ExperimentTaskStatus.SUCCESS;

            }
        }

//        public class WaitForExecuteServicesState extends State {
//
//            protected WaitForExecuteServicesState() {
//                super("Waiting for execute-services to finish on " + node.getClientId());
//            }
//
//            @Nonnull
//            @Override
//            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
//                try {
//                    for (FXExecuteService executeService : node.getExecuteServices()) {
//                        if (executeService.getFinishedFile() == null)
//                            continue;
//
//                        boolean fileFound = false;
//                        int tries = 0;
//
//                        while (!fileFound && tries < MAX_EXECUTE_FINISHED_FILE_DETECTION_TRIES) {
//                            Session.Command command =
//                                    ssh.startSession().exec("test -f " + executeService.getFinishedFile());
//                            command.join();
//
//                            fileFound = command.getExitStatus() == 0;
//
//                            tries++;
//                            if (!fileFound) {
//                                updateMessage("Looked for '" + executeService.getFinishedFile() + "' " + tries + " times.");
//                                Thread.sleep(TIME_BETWEEN_FINISHED_FILE_DETECTION_TRIES_MS);
//                            }
//                        }
//
//                        if (!fileFound) {
//                            updateMessage("Timed out while waiting for finished-file '" + executeService.getFinishedFile() + "' to appear");
//                            return ExperimentTaskStatus.FAILED;
//                        } else {
//                            updateMessage("Found finished-file '" + executeService.getFinishedFile() + "'");
//                        }
//                    }
//
//                    updateMessage("Found all finished-files");
//                    return ExperimentTaskStatus.SUCCESS;
//
//
//                } catch (IOException ex) {
//                    LOG.error("Error while waiting for finished-files on {}", node.getClientId(), ex);
//                    updateMessage("Error while waiting for finished-files on {}" + ex.getMessage());
//                    return ExperimentTaskStatus.FAILED;
//                }
//            }
//        }

        public class UploadState extends State {
            protected UploadState() {
                super("Uploading files on " + node.getClientId());
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                if (ssh == null)
                    throw new IllegalStateException("ssh == null");
                if (externalFileUtil == null)
                    throw new IllegalStateException("externalFileUtil == null");

                //We need to wait for all files to be ready
                //noinspection ConstantConditions
                job.getExperiment().getExperimentSpecificationFileManager().waitUntilReady();

                //uploads file AND scripts (but don't execute scripts yet)
                List<UploadLikeSpec> uploads = new ArrayList<>(experimentSpecification.getUploads());
                uploads.addAll(experimentSpecification.getExecutes());

                for (UploadLikeSpec uploadSpec : uploads) {
                    if (uploadSpec.getNodes() == null || uploadSpec.getNodes().contains(clientId)) {
                        byte[] content = job.getExperiment().getExperimentSpecificationFileManager().getBytes(uploadSpec.getSource());

                        assert externalFileUtil != null;
                        //TODO: use dir with "scripts" or "uploads" type as appropriate, fall back to homedir
                        String remotePath = uploadSpec.getPath() == null ?
                                externalFileUtil.getHomeDir()+"/"+uploadSpec.getSource().getBasename():
                                uploadSpec.getPath();
                        String name = externalFileUtil.getFilename(remotePath);

                        try {
                            LOG.debug("ExperimentSpecification.upload is uploading {} ({} bytes) to {} on {}",
                                    name, content.length, remotePath, node.getClientId());
                            ssh.newSCPFileTransfer().upload(
                                    new ByteArraySshjFile(name, content, uploadSpec.getPermissions()),
                                    remotePath);

                            remotePathByUploadLikeSpec.put(uploadSpec, remotePath);
                        } catch (IOException ex) {
                            LOG.error("Error while uploading {} to {} on {}", name, remotePath, node.getClientId(), ex);
                            updateMessage(String.format("Error while uploading %s to %s on %s: %s", name, remotePath, node.getClientId(), ex.getMessage()));
                            return ExperimentTaskStatus.FAILED;
                        }
                    }
                }

                return ExperimentTaskStatus.SUCCESS;
            }
        }

//        public class ExecuteState extends State {
//
//            protected ExecuteState() {
//                super("Executing ansible-playbook '" + ansibleService.getExecutePlaybook() + "' on " + node.getClientId());
//
//            }
//
//            @Nonnull
//            @Override
//            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
//
//                try {
//                    //first install the requirements if needed
//                    if (ansibleService.getInstallRequirements() != null) {
//                        int dividerPosition = ansibleService.getInstallRequirements().lastIndexOf('/');
//                        String folder = ansibleService.getInstallRequirements().substring(0, dividerPosition);
//                        String filename = ansibleService.getInstallRequirements().substring(dividerPosition + 1);
//
//                        String galaxyCommand = ansibleService.getGalaxyCommand();
//                        if (galaxyCommand == null)
//                            galaxyCommand = "ansible-galaxy";
//
//                        String galaxyCommandString = String.format("cd %s && %s install -r %s -p roles",
//                                folder, galaxyCommand, filename);
//                        LOG.debug("Executing galaxy command: "+galaxyCommandString);
//                        Session.Command c =
//                                ssh.startSession().exec(galaxyCommandString);
//
//                        new InputStreamToStringPropertyReaderThread(c.getInputStream(), galaxyInputStream, Level.INFO, "GALAXY-OUTPUT: ").start();
//                        new InputStreamToStringPropertyReaderThread(c.getErrorStream(), galaxyErrorStream, Level.WARN, "GALAXY-ERROR: ").start();
//
//                        c.join();
//
//                        if (c.getExitStatus() != 0) {
//                            updateMessage("ansible-galaxy execution was unsuccessful. Aborting");
//                            return ExperimentTaskStatus.FAILED;
//                        }
//                    }
//
//                    assert ansibleService.getExecutePlaybook() != null;
//                    int dividerPosition = ansibleService.getExecutePlaybook().lastIndexOf('/');
//                    String folder = ansibleService.getExecutePlaybook().substring(0, dividerPosition);
//                    String filename = ansibleService.getExecutePlaybook().substring(dividerPosition + 1);
//
//                    String playbookCommand = ansibleService.getPlaybookCommand();
//                    if (playbookCommand == null)
//                        playbookCommand = "ansible-playbook";
//
//                    String arguments = "";
//                    arguments += filename;
//
//                    if (ansibleService.getDebug() == Boolean.TRUE)
//                        arguments += " -v";
//
//                    if (ansibleService.getInventory() != null)
//                        arguments += " -i '"+ansibleService.getInventory()+"'";
//
//                    //this must be the LAST argument, as it is a redirect and pipe
//                    if (ansibleService.getLogfile() != null)
//                        arguments += " 2>&1 | tee '"+ansibleService.getLogfile()+"'";
//
//                    String playbookCommandString = String.format("cd %s && %s %s", folder, playbookCommand, arguments);
//                    LOG.debug("Executing playbook command: "+playbookCommandString);
//                    Session.Command c =
//                            ssh.startSession().exec(playbookCommandString);
//
//                    new InputStreamToStringPropertyReaderThread(c.getInputStream(), playbookInputStream, Level.INFO, "PLAYBOOK-OUTPUT: ").start();
//                    new InputStreamToStringPropertyReaderThread(c.getErrorStream(), playbookErrorStream, Level.WARN, "PLAYBOOK-ERROR: ").start();
//
//                    c.join();
//
//
//                    if (c.getExitStatus() != 0) {
//                        updateMessage("ansible-playbook execution was unsuccessful.");
//                        return ExperimentTaskStatus.FAILED;
//                    }
//
//
//                } catch (TransportException | ConnectionException ex) {
//                    throw new JFedException("SSH-exception while executing ansible: " + ex.getMessage(), ex);
//                }
//
//                return ExperimentTaskStatus.SUCCESS;
//            }
//
//            public FXRspecNode getNode() {
//                return node;
//            }
//        }
    }
}
